//
// UR5FFR Si5351 PICO NWT
// v 1.3
// Copyright (c) Andrey Belokon, 2020 Odessa 
// http://dspview.com
//

#include <avr/eeprom.h> 
#include "i2c.h"
#include "si5351a.h"

/*
  I2C address mapping
  I2C device found at address 0x60  ! Si5351
*/

#define SI5351_CALIBRATION       25003050

#define XTAL_MIN  24000000
#define XTAL_MAX  31000000
long EEMEM XTAL_FREQ_EEMEM = 0;

#define FIRMWARE_VERSION    "1.3;PICO"

Si5351 vfo5351;

char *ReadNumber(char *p, long *value)
{
  long val = 0;
  char c;
  bool negate;
  //if (negate = (*p == '-')) p++;
  while ((c = *p++) && isdigit(c)) val = val*10+c-'0';
  *value = val;//(negate ? -val : val);
  return p;
}

char *SkipWhitespace(char *p){
  while (*p && (*p == ' ' || *p == ','))
    p++;
  return p;
} 

char cmd_buf[50];
int icmd_buf = 0;
long freq;

#define AF_FREQ   1000
#define SAMPLING_K   16
#define SAMPLING_FREQ   (SAMPLING_K*AF_FREQ)

// уровень сигнала на выходе Si5351. 0=2mA, 1=4mA, 2=6mA, 3=8mA
/*
 * CLK0 - OUT
 * CLK1 - SA612
 * CLK2 - not used
 */
#define SI5351_CLK0_DRIVE   3
#define SI5351_CLK1_DRIVE   3
#define SI5351_CLK2_DRIVE   0

bool UpdateFreq()
{
  return vfo5351.set_freq(
    // по миниуму шумовой дорожки
    (freq < 130000000 ? freq+AF_FREQ : freq-AF_FREQ),
    freq,
    0
  );
}

bool IsFreqOK()
{
  return vfo5351.is_freq_ok(0) && vfo5351.is_freq_ok(1);
}

EMPTY_INTERRUPT (TIMER1_COMPB_vect);

// 0 - find min
// 1 - find max
// 2 - measure
// 3 - stop
volatile byte measure_stage = 0;
volatile int last_val;
volatile int nstep = 0;

// 48 will give you a bandwidth around 186 hz            //
// 64 will give you a bandwidth around 140 hz            //
// 96 will give you a bandwidth around 94 hz             //
// 128 will give you a bandwidth around 70 hz            //
#define N 256
float coeff;
volatile float Q0,Q1,Q2;

// http://gammon.com.au/adc
// https://arduino.stackexchange.com/questions/45927/arduino-continuously-reading-adc-value-using-interrupt
// https://github.com/krisdover/FastADC

ISR (ADC_vect)
{
  int16_t val = ADC;
  switch (measure_stage) {
    case 0:
        if (val <= last_val) last_val = val;
        else measure_stage = 1;
        if (++nstep >= SAMPLING_K) {
          nstep = 0;
          measure_stage = 2;
        }
        break;
        
    case 1:  
        if (val >= last_val) {
          last_val = val;
          if (++nstep < SAMPLING_K) 
            break;
        }
        nstep = 0;
        measure_stage = 2;
        
    case 2:  
        Q0 = coeff * Q1 - Q2 + val;
        Q2 = Q1;
        Q1 = Q0;  
        if (++nstep >= N) 
          measure_stage = 3;
        break;
        
    case 3:  
        break;
  }
} 

float ReadData()
{
  nstep = 0;
  last_val=10000;
  Q1=Q2=0;
  measure_stage = 0;

  // TIMER 1 for interrupt frequency 16000 Hz:
  TCCR1A = 0; // set entire TCCR1A register to 0
  TCCR1B = 0; // same for TCCR1B
  TCNT1  = 0; // initialize counter value to 0
  // set compare match register for 16000 Hz increments
  OCR1A = 999; // ?????????
  OCR1B = 999; // = 16000000 / (1 * 16000) - 1 (must be <65536)
  // turn on CTC mode, 1 prescaler
  TCCR1B = bit(WGM12) | bit(CS10);
  // enable timer compare interrupt
  TIMSK1 = bit(OCIE1B);
  
  // http://yaab-arduino.blogspot.com/2015/02/fast-sampling-from-analog-input.html
  // http://gammon.com.au/adc
  // set an ADC prescale to 16 --> 76.8kHz
  // 16 have best precision/speed ratio
  // Prescaler of 16
  ADCSRA =  bit(ADEN) | bit(ADIE) | bit(ADIF) | bit(ADPS2);   // turn ADC on, want interrupt on completion
  ADMUX = 0; // A0 & ext ref
  ADCSRB = bit(ADTS0) | bit(ADTS2);  // Timer/Counter1 Compare Match B
  ADCSRA |= bit(ADATE);   // turn on automatic triggering

  while (measure_stage < 3) { /* wait */ };
  
  // stop timer & disable timer compare interrupt
  TCCR1B &= ~((1 << CS12) | (1 << CS11) | (1 << CS10));
  TIMSK1 &= ~(1 << OCIE1B);
  ADCSRA =  bit(ADEN) | bit(ADPS2);   // turn ADC on, Prescaler of 16, no interrupt

  // we do only need the real part //
  // Q1*Q1 + Q2*Q2 - Q1*Q2*coeff
  return Q1*Q1 + Q2*(Q2-Q1*coeff); // mul optimisation
}

void ParseCommand()
{
  char *p = cmd_buf;
  char cmd = *p++;
  long step,count,xfreq;
  p = SkipWhitespace(p);
  switch (cmd) {
    case 'w':
      // get firmware version
      Serial.println(FIRMWARE_VERSION);
      break;
      
    case 'c':
      // калибровка частоты сишки
      vfo5351.set_freq(5000000,5000000,5000000);
      vfo5351.out_calibrate_freq();
      break;
      
    case 'x':
      // задать калибровочную частоту
      // x 25003566
      ReadNumber(p,&xfreq);
      if (xfreq >= XTAL_MIN && xfreq <= XTAL_MAX) {
        eeprom_write_block(&xfreq, &XTAL_FREQ_EEMEM, sizeof(xfreq));
        vfo5351.set_xtal_freq(xfreq);
      }
      break;
      
    case 'o':
      // off. отключить генерацию
      vfo5351.set_freq(0,0,0);
      break;
      
    case 'f':
      // установить частоту
      // f 5000000
      ReadNumber(p,&freq);
      if (IsFreqOK())
        UpdateFreq();
      else {
        UpdateFreq();
        delay(100);
      }
      break;
      
    case 'r':
      // read data
      // r step count
      p = ReadNumber(p,&step);
      p = SkipWhitespace(p);
      ReadNumber(p,&count);
      while (count-- > 0) {
        if (UpdateFreq()) delay(50);
        else delay(1);
        if (IsFreqOK()) {
          Serial.println(ReadData());
        } else {
          Serial.println("*");
        }
        freq += step;
      }
      break;
      
  }
  icmd_buf = 0;
}

void loop()
{
  if (Serial.available()) {
    int inbyte = Serial.read();
  
    if (inbyte == 13){
      cmd_buf[icmd_buf] = 0;
      icmd_buf = 0;
      ParseCommand();
    } else if (icmd_buf < sizeof(cmd_buf)-1) {
      cmd_buf[icmd_buf++] = inbyte;
    }
  }
}

void setup()
{
  long xtal;
  eeprom_read_block(&xtal, &XTAL_FREQ_EEMEM, sizeof(xtal));
  if (xtal < XTAL_MIN || xtal > XTAL_MAX)
    xtal = SI5351_CALIBRATION;

  Serial.begin(115200);
  Serial.println(FIRMWARE_VERSION);
  
  // The basic goertzel calculation
  int  k = (int) (0.5 + ((N * 1.0 * AF_FREQ) / SAMPLING_FREQ));
  coeff = 2.0 * cos((2.0 * PI * k) / N);
  
  i2c_init();
  vfo5351.VCOFreq_Max = 800000000; // для использования "кривых" SI-шек с нестабильной генерацией
  vfo5351.setup(
    SI5351_CLK0_DRIVE,
    SI5351_CLK1_DRIVE,
    SI5351_CLK2_DRIVE
  ); 
  vfo5351.set_xtal_freq(xtal);

  pinMode(A0,INPUT);

  // debug
  freq = 3000000;
  UpdateFreq();
}
